"""
3D AnnotatorWidget Widget
"""
import argparse
import json
import os
from pathlib import Path

import napari
import numpy as np
import pandas as pd
from magicgui import magicgui
from napari.utils.notifications import show_info
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QMessageBox, QLabel, QSlider

from ._annotator import Annotator
from ._params import Params
from .utils.io import annotation_to_pandas, pandas_to_annotations

TEXT_PROP = {
    'text': 'label',
    'anchor': 'upper_left',
    'translation': [2, 2, 5],
    'size': 7,
    'color': 'green'
}


class AnnotatorWidget(QWidget):
    """
    Annotator Widget
    """

    def __init__(self, napari_viewer):
        super().__init__()
        self.viewer = napari_viewer
        self.annotation_layer = None
        image_layer = self.get_image_layer()
        self.image = image_layer.data if image_layer is not None else None
        path = image_layer.source.path if image_layer is not None else None
        self.datapath = os.path.dirname(path) if path is not None else '.'
        self.filename = path[:-len(path.split('.')[-1]) - 1] + '.csv' if path is not None else 'annotations.csv'
        self.param_filename = os.path.join(self.datapath, 'params.json') if path is not None else 'params.json'
        self.params = Params()
        self.set_params()
        self.setup_ui()

    def set_params(self):
        """Set annotation parameters"""
        self.voxel_params()
        self.sigma_param()
        self.display_params()
        self.ac_parameters1()
        self.ac_parameters2()

    def voxel_params(self, voxel_size_xy: float = 0.1, voxel_size_z: float = 0.1):
        """
        Specify voxel size.

        Parameters
        ----------
        voxel_size_xy : float
            Voxel size in xy (microns)
        voxel_size_z : float
            Voxel size in z (microns)
        """
        self._set_scale([voxel_size_z, voxel_size_xy, voxel_size_xy])
        self.params.set_scale(self.scale)

    def sigma_param(self, sigma_um: float = 0.2):
        """
        Specify voxel size.

        Parameters
        ----------
        sigma_um : flaot
            Gaussian sigma (in microns) to smooth the image for active contour refinement.
        """
        self.params.set_smoothing(sigma_um)

    def display_params(self, line_width: float = 0.5):
        """

        Parameters
        ----------
        line_width : float
            Width of the annotation lines in the viewer.
        """
        self.params.set_linewidth(line_width)

    def ac_parameters1(self, alpha: float = 0.01, beta: float = 0.1, gamma: float = 1):
        """

        Parameters
        ----------
        alpha : float
            Active contour weight for the first derivative
        beta : float
            Active contour weight for the second derivative
        gamma : float
            Active contour weight for the image contribution
        """
        self.params.set_coef(alpha=alpha, beta=beta, gamma=gamma)

    def ac_parameters2(self, n_iter: int = 1000, n_interp: int = 3, end_coef: float = 0.0):
        """

        Parameters
        ----------
        n_iter : int
            Number of iterations of the active contour
            Width of the annotation lines in the viewer.
        n_interp : int
            Number of points to interpolate between each pair of annotated point
        end_coef : float
            Coefficient (between 0 and 1) to scale the forces applied to the contour end points.
            Set to 0 to fix the end points.
        """
        self.params.set_ac_parameters(n_iter=n_iter, n_interp=n_interp,
                                      end_coef=end_coef)

    def load_annotations(self, filename=Path('.')):
        """

        Parameters
        ----------
        filename : Path
            Filename to load annoations

        """
        df = pd.read_csv(filename)
        data, labels = pandas_to_annotations(df)
        self.viewer.add_shapes(data, name='existing_annotations',
                               shape_type='path', edge_color='green',
                               edge_width=self.params.line_width,
                               scale=self.viewer.layers[0].scale,
                               blending='additive',
                               properties={'label': labels}, text=TEXT_PROP)
        if not self.annotation_layer_exists():
            self.add_annotation_layer()
        self.annotation_layer.add(data, shape_type='path', edge_color='green',
                                  edge_width=self.params.line_width)

    def load_parameters(self, filename=Path('.')):
        """

        Parameters
        ----------
        filename : Path
            Filename to load parameters

        """
        with open(filename, 'r') as f:
            params = json.load(f)
        params = argparse.Namespace(**params)
        self.magic_voxel_params.voxel_size_xy.value = params.voxel_size_xy
        self.magic_voxel_params.voxel_size_z.value = params.voxel_size_z
        self.magic_sigma_param.sigma_um.value = params.sigma_um
        self.magic_display_params.line_width.value = params.line_width
        self.magic_ac_parameters1.alpha.value = params.alpha
        self.magic_ac_parameters1.beta.value = params.beta
        self.magic_ac_parameters1.gamma.value = params.gamma
        self.magic_ac_parameters2.n_iter.value = params.n_iter
        self.magic_ac_parameters2.n_interp.value = params.n_interp
        self.magic_ac_parameters2.end_coef.value = params.end_coef

    def get_param_filename(self, filename=Path('.')):
        """

        Parameters
        ----------
        filename : Path
            Filename to save parameters
        """
        self.param_filename = filename
        self.save_parameters()

    def save_parameters(self):
        self.params.save(self.param_filename)
        print(rf"Saved to: {self.param_filename}")

    def get_annotation_filename(self, filename=Path('.')):
        """

        Parameters
        ----------
        filename : Path
            Filename to save annotations
        """
        self.filename = filename
        self.save_annotations()

    def save_annotations(self):
        if self.annotation_layer is not None and self.annotation_layer.nshapes > 1:
            annotation_to_pandas(self.annotation_layer.data[1:]).to_csv(self.filename, index=False)
        else:
            pd.DataFrame().to_csv(self.filename, index=False)
        print(rf"Saved to: {self.filename}")

    def set_maxval(self):
        maxval = self.sld.value()
        img_layer = self.get_image_layer()
        if img_layer is not None:
            if self.image is None:
                self.image = img_layer.data
                self.sld.setMaximum(self.image.max())
            img_layer.data = np.where(self.image > maxval, 0, self.image)

    def get_image_layer(self):
        if len(self.viewer.layers) > 0 and isinstance(self.viewer.layers[0], napari.layers.Image):
            return self.viewer.layers[0]
        else:
            return None

    def annotation_layer_exists(self):
        """
        Check if the annotation layer still exists in the viewer.
        The self.annotation_layer variable will not get updated if the layer is deleted from the viewer.
        """
        return len([layer for layer in self.viewer.layers
                    if isinstance(layer, napari.layers.Shapes) and layer.name == 'annotations']) > 0

    def add_annotation_layer(self):
        """
        Add an annotation layer to the napari viewer.
        """
        img_layer = self.get_image_layer()
        self.viewer.dims.ndisplay = 3
        if img_layer is not None:
            if self.annotation_layer_exists():
                answer = self._confirm_adding_second_layer()
                if answer == QMessageBox.No:
                    return
            if len(img_layer.data.shape) == 3:
                self.annotator = Annotator(self.viewer, img_layer, self.params)
                self.annotation_layer = self.annotator.annotation_layer
            else:
                show_info("Only 3D gray-scale images are currently supported! "
                          rf"The current image has {len(img_layer.data.shape)} dimensions.")
        else:
            show_info("No images open! Please open an image first")

    def setup_ui(self):
        layout = QVBoxLayout()
        self.setLayout(layout)

        # Load parameters from file
        self._add_magic_function(magicgui(self.load_parameters, layout='vertical', auto_call=True,
                                          filename={"mode": "r",
                                                    "label": "Load existing parameters:",
                                                    "filter": "*.json",
                                                    "value": self.param_filename}),
                                 layout)

        # Voxe size and annotation parameters
        l1 = QHBoxLayout()
        l1.addWidget(QLabel("Image parameters"))
        layout.addLayout(l1)
        self.magic_sigma_param = magicgui(self.sigma_param, layout='vertical', auto_call=True)
        self._add_magic_function(self.magic_sigma_param, l1)
        self.magic_voxel_params = magicgui(self.voxel_params, layout='vertical', auto_call=True)
        self._add_magic_function(self.magic_voxel_params, layout)

        # Slider for masking out spindle
        l2 = QHBoxLayout()
        layout.addLayout(l2)
        l2.addWidget(QLabel("Mask out bright pixels"))
        self.sld = QSlider(Qt.Horizontal)
        self.sld.valueChanged.connect(self.set_maxval)
        img_layer = self.get_image_layer()
        if img_layer is not None:
            self.sld.setMaximum(img_layer.data.max())
            self.sld.setValue(img_layer.data.max())
        self.sld.setValue(self.sld.maximum())
        l2.addWidget(self.sld, Qt.Horizontal)
        self.set_maxval()

        # "Add annotation layer" button
        btn = QPushButton("Add annotation layer")
        btn.clicked.connect(self.add_annotation_layer)
        layout.addWidget(btn)

        # Display parameters
        l3 = QHBoxLayout()
        layout.addLayout(l3)
        l3.addWidget(QLabel("Display parameters"))
        self.magic_display_params = magicgui(self.display_params, layout='vertical', auto_call=True)
        self._add_magic_function(self.magic_display_params, l3)

        # Active contour parameters
        layout.addWidget(QLabel("Active contour parameters"))
        l4 = QHBoxLayout()
        layout.addLayout(l4)

        self.magic_ac_parameters1 = magicgui(self.ac_parameters1, layout='vertical', auto_call=True)
        self.magic_ac_parameters2 = magicgui(self.ac_parameters2, layout='vertical', auto_call=True)
        self._add_magic_function(self.magic_ac_parameters1, l4)
        self._add_magic_function(self.magic_ac_parameters2, l4)

        # Save parameters
        l5 = QHBoxLayout()
        layout.addLayout(l5)
        self._add_magic_function(magicgui(self.get_param_filename, layout='vertical', auto_call=True,
                                          filename={"mode": "w",
                                                    "label": "Save parameters:",
                                                    "filter": "*.json",
                                                    "value": self.param_filename}),
                                 l5)
        btn_save = QPushButton("Save parameters")
        btn_save.clicked.connect(self.save_parameters)
        l5.addWidget(btn_save)

        # Load annotations
        l6 = QHBoxLayout()
        layout.addLayout(l6)
        self._add_magic_function(magicgui(self.load_annotations, layout='vertical', auto_call=True,
                                          filename={"mode": "r",
                                                    "label": "Load existing annotations:",
                                                    "filter": "*.csv",
                                                    "value": self.datapath}),
                                 l6)

        # Save annotations
        l7 = QHBoxLayout()
        layout.addLayout(l7)
        self._add_magic_function(magicgui(self.get_annotation_filename, layout='vertical', auto_call=True,
                                          filename={"mode": "w",
                                                    "label": "Save annotations:",
                                                    "filter": "*.csv",
                                                    "value": self.filename}),
                                 l7)
        btn_save = QPushButton("Save annotations")
        btn_save.clicked.connect(self.save_annotations)
        l7.addWidget(btn_save)

    def _set_scale(self, scale):
        self.scale = scale
        if np.min(scale) > 0:
            for i in range(len(self.viewer.layers)):
                self.viewer.layers[i].scale = scale
            self.viewer.dims.ndisplay = 2
            self.viewer.dims.ndisplay = 3

    def _add_magic_function(self, function, _layout):
        # self.viewer.layers.events.inserted.connect(function.reset_choices)
        # self.viewer.layers.events.removed.connect(function.reset_choices)
        _layout.addWidget(function.native)

    def _confirm_adding_second_layer(self):
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Question)

        msg.setWindowTitle("Annotation layer already exists")
        msg.setText("Annotation layer already exists! Are you sure you want to add another annotation layer?")
        msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        return msg.exec_()



import json
import os

import numpy as np
import pandas as pd
import pytest
from napari_filament_annotator import AnnotatorWidget
from napari_filament_annotator.utils.const import COLS
from napari_filament_annotator.utils.io import annotation_to_pandas


# make_napari_viewer is a pytest fixture that returns a napari viewer object
# capsys is a pytest fixture that captures stdout and stderr output streams
@pytest.fixture
def annotator_widget(make_napari_viewer, capsys):
    # make viewer and add an image layer using our fixture
    viewer = make_napari_viewer()
    widget = AnnotatorWidget(viewer)
    widget.add_annotation_layer()
    assert widget.annotation_layer_exists() is False
    return widget


@pytest.fixture
def annotator_widget_with_image(annotator_widget):
    annotator_widget.viewer.add_image(np.random.randint(0, 100, (50, 100, 100)))
    return annotator_widget


@pytest.fixture
def annotator(annotator_widget_with_image):
    annotator_widget_with_image.add_annotation_layer()
    assert annotator_widget_with_image.annotation_layer_exists()
    return annotator_widget_with_image.annotator


def test_time_series(annotator_widget):
    annotator_widget.viewer.add_image(np.random.randint(0, 100, (5, 50, 100, 100)))
    annotator_widget.add_annotation_layer()
    assert annotator_widget.annotation_layer_exists() is False


def test_add_delete(annotator, polygons):
    layer = annotator.annotation_layer
    assert layer.nshapes == 1

    annotator.near_points = polygons[0][0].copy()
    annotator.far_points = polygons[0][1].copy()
    annotator.draw_polygon(layer)
    assert layer.nshapes == 2

    annotator.delete_the_last_point(layer)
    assert layer.nshapes == 2
    assert len(polygons[0][0]) == len(annotator.near_points) + 1 == len(annotator.far_points) + 1

    annotator.delete_the_last_shape(layer)
    assert layer.nshapes == 1


def test_intersection(annotator, polygons):
    layer = annotator.annotation_layer
    annotator.near_points = polygons[0][0].copy()
    annotator.far_points = polygons[0][1].copy()
    assert len(annotator.near_points) > 0
    annotator.draw_polygon(layer)
    annotator.calculate_intersection(layer)
    assert layer.nshapes == 2
    assert len(annotator.polygons) == 1
    assert len(annotator.near_points) == len(annotator.far_points) == 0

    annotator.near_points = polygons[1][0].copy()
    annotator.far_points = polygons[1][1].copy()
    annotator.draw_polygon(layer)
    annotator.calculate_intersection(layer)
    assert layer.nshapes == 2
    assert len(annotator.near_points) == len(annotator.far_points) == len(annotator.polygons) == 0
    points = layer.data[-1].copy()
    annotator.delete_the_last_filament_point(layer)
    assert (points[:-1] == layer.data[-1]).all()
    annotator.delete_the_first_filament_point(layer)
    assert (points[1:-1] == layer.data[-1]).all()


def test_params(annotator, tmp_path):
    # test that all parameters are set
    params = ['scale', 'sigma', 'line_width', 'alpha',
              'beta', 'gamma', 'n_iter', 'n_interp', 'end_coef']
    for param in params:
        assert param in vars(annotator.params)
    assert len(annotator.params.scale) == len(annotator.params.sigma) == 3
    annotator.params.save(os.path.join(tmp_path, 'params.json'))
    assert os.path.exists(os.path.join(tmp_path, 'params.json'))


def test_add_annotation_layer(annotator_widget):
    # call "add annotation layer" without image and assert no shape layers were added
    annotator_widget.add_annotation_layer()
    assert annotator_widget.annotation_layer_exists() is False

    # add an image
    annotator_widget.viewer.add_image(np.random.randint(0, 100, (50, 100, 100)))

    # call "add annotation layer" again and assert a new shape layer was added
    annotator_widget.add_annotation_layer()
    assert annotator_widget.annotation_layer_exists()


def test_maxval(annotator_widget_with_image):
    annotator_widget_with_image.sld.setValue(50)
    assert annotator_widget_with_image.get_image_layer().data.max() == 50


def test_io(annotator_widget_with_image, tmp_path, paths):
    fn = os.path.join(tmp_path, 'annotations.csv')
    df = annotation_to_pandas(paths)
    df.to_csv(fn)
    annotator_widget_with_image.load_annotations(fn)
    assert annotator_widget_with_image.annotation_layer_exists()
    assert len(annotator_widget_with_image.annotation_layer.data) - 1 == len(paths)

    annotator_widget_with_image.get_annotation_filename(fn)
    df2 = pd.read_csv(fn)
    assert (df[COLS].values == df2[COLS].values).all()


def test_param_io(annotator_widget, tmp_path):
    fn = os.path.join(tmp_path, 'params.json')
    annotator_widget.get_param_filename(fn)
    assert os.path.exists(fn)
    with open(fn, 'r') as f:
        params = json.load(f)
    params['voxel_size_xy'] = 2
    with open(fn, 'w') as f:
        json.dump(params, f)
    annotator_widget.load_parameters(fn)
    assert annotator_widget.params.scale[-1] == 2



import numpy as np
from napari_filament_annotator import load_sample_image


def test_sample_data():
    img = load_sample_image()
    assert isinstance(img, list)
    assert isinstance(img[0], tuple)
    assert isinstance(img[0][0], np.ndarray)
    assert len(img[0][0].shape) == 3



import numpy as np
from skimage.filters import sobel


def gradient(img, spacing=None):
    """
    Calculate image gradient along all axes.

    Parameters
    ----------
    img : np.ndarray
        Input 3D image
    spacing : tuple, list or array
        Voxel size, (z, y, x).

    Returns
    -------
    list of np.ndarray:
        Image gradients along all three axes.
        List of length 3, each element having the same shape as the input image.
    """
    if spacing is None:
        spacing = np.ones(img.ndim)
    grad = [sobel(img, axis=i) / spacing[i] for i in range(img.ndim)]
    return grad


def snap_to_bright(snake, img=None, grad=None, spacing=None,
                   alpha=0.01, beta=0.1, gamma=1,
                   n_iter=1000, end_coef=0.01, n_interp=5, **_):
    """
    Snap the annotation to the brightest intensity and regularize the curve based on active contours.

    Parameters
    ----------
    snake : np.ndarray
        N x 3 array of the filament coordinates.
    img : np.ndarray
        3D input image (after smoothing, if applicable)
    grad : list of np.ndarray
        Image gradients along all three axes.
        List of length 3, each element having the same shape as the input image.
    spacing : tuple, list or array
        Voxel size, (z, y, x).
    alpha : float
        Active contour weight for the first derivative.
    beta : float
        Active contour weight for the seconf derivative.
    gamma : float
        Active contour weight for the image contribution.
    n_iter : int
        Number of iterations of the active contour.
        Width of the annotation lines in the viewer.
    end_coef : float
        Coefficient (between 0 and 1) to scale the forces applied to the contour end points.
        Set to 0 to fix the end points.
    n_interp : int
        Number of points to interpolate between each annotated point.

    Returns
    -------
    np.ndarray
        M x 3 array of regularized filament coordinates.
        M is approximately equal to N * n_interp

    """
    if spacing is None:
        spacing = np.ones(3)
    if grad is None:
        if img is None:
            raise ValueError("Either image or gradient must be provided!")
        grad = gradient(img, spacing)

    if not isinstance(snake, np.ndarray) or len(snake.shape) != 2 or snake.shape[1] != 3:
        raise ValueError("Input snake must be a numpy array of shape N x 3")

    snake = _interpolate(snake, npoints=n_interp)  # interpolate between the points
    snake = _evolve_snake(snake, n_iter, grad, spacing, alpha, beta, gamma, end_coef)  # evolve the snake

    return snake


def _evolve_snake(snake, n_iter, grad, spacing, alpha, beta, gamma, end_coef):
    coef = np.ones_like(snake)  # coefficient to weight end points vs all other points
    coef[0] = end_coef
    coef[-1] = end_coef
    c = np.arange(1, n_iter + 1)[::-1] / n_iter  # weight for each iteration; linearly decrease with each iteration
    for it in range(n_iter):
        coord = tuple(np.int_(np.round_(snake)).transpose())

        # normalize image force, based on the gradient at the current coordinate
        fimg = np.array([grad[i][coord] for i in range(len(grad))])
        fimg = - fimg.transpose()
        fimg = fimg / fimg.max()

        # normalized snake (internal) force
        d2, d4 = _get_derivatives_2_4(snake * spacing)
        fsnake = d2 * alpha + d4 * beta
        fsnake = fsnake / np.max(fsnake)

        # update the snake with the final force
        snake = snake - c[it] * coef * (fimg * gamma + fsnake) / (gamma + 1)
        # make sure snake coordinates are not outside image
        snake = _fit_to_image_shape(snake, grad[0].shape)
    return snake


def _fit_to_image_shape(snake, shape):
    shape = np.array(shape) - 1
    snake = np.array([np.max([[0, 0, 0], np.min([shape, snake[i]], axis=0)], axis=0) for i in range(len(snake))])
    return snake


def _interpolate(x, npoints=5):
    if npoints > 1:
        new_x = []
        for i in range(1, len(x)):
            new_x.append(np.linspace(x[i - 1], x[i], npoints, endpoint=False))
        new_x.append(x[-1:])
        new_x = np.concatenate(new_x, axis=0)
        return new_x
    else:
        return x


def _remove_corners(x, k=3):
    # first and second derivatives
    d1, d2 = _get_derivatives_1_2(x)
    d1 = np.sqrt((d1 ** 2).sum(1))
    d2 = np.sqrt((d2 ** 2).sum(1))

    # only consider points where the second derivative is greater than the first
    d2 = np.where(d2 / d1 > 1, d2, 0)

    # find the maximum of the second derivative (the corners) among the first and the last k points
    i1 = np.argmax(d2[:k])
    i2 = np.argmax(d2[-k:][::-1])

    # remove points before and after these maxima
    x = x[i1:]
    if i2 > 0:
        x = x[:-i2]
    return x


def _get_derivatives_1_2(x):
    x = np.concatenate([x[1:3][::-1], x, x[-3:-1][::-1]])
    d1 = x - np.roll(x, -1, 0)
    d2 = np.roll(x, 1, 0) + np.roll(x, -1, 0) - 2 * x
    return d1[2:-2], d2[2:-2]


def _get_derivatives_2_4(x):
    x = np.concatenate([x[1:3][::-1], x, x[-3:-1][::-1]])
    d2 = np.roll(x, 1, 0) + np.roll(x, -1, 0) - 2 * x
    d4 = np.roll(x, 2, 0) - 4 * np.roll(x, 1, 0) + 6 * x - 4 * np.roll(x, -1, 0) + np.roll(x, -2, 0)
    return d2[2:-2], d4[2:-2]



import numpy as np
import pytest
from napari_filament_annotator.utils.postproc import snap_to_bright, gradient


@pytest.fixture
def grad(img_snake):
    return gradient(img_snake[0])


@pytest.fixture
def init_snake(img_snake):
    _, snake = img_snake
    init = snake.copy()
    init[1:-1] = init[1:-1] + (np.random.rand(*init[1:-1].shape) - 0.5) * np.array([2, 10, 10])
    return snake, init


def test_snapping(img_snake, init_snake, grad):
    img, _ = img_snake
    snake, init = init_snake
    snake2 = snap_to_bright(init, img=img, alpha=0.01, beta=0.1, gamma=1,
                            n_iter=100, end_coef=0, n_interp=0, remove_corners=False)
    snake3 = snap_to_bright(init, grad=grad, alpha=0.01, beta=0.1, gamma=1,
                            n_iter=100, end_coef=0, n_interp=0, remove_corners=False)
    assert len(snake) == len(snake2) == len(snake3)
    assert (snake[0] == snake2[0]).all()
    assert (snake[0] == snake3[0]).all()
    assert (snake[-1] == snake2[-1]).all()
    assert (snake[-1] == snake3[-1]).all()
    assert (snake2 == snake3).all()
    dist = np.mean(np.sqrt(np.sum((snake - snake2) ** 2, axis=1)))
    assert dist < 3


def test_snapping_interp(init_snake, grad):
    snake, init = init_snake
    snake2 = snap_to_bright(init, grad=grad, alpha=0.01, beta=0.1, gamma=1,
                            n_iter=100, end_coef=0, n_interp=5, remove_corners=True)
    assert len(snake2) == (len(snake) - 1) * 5 + 1
    assert (snake[0] == snake2[0]).all()
    assert (snake[-1] == snake2[-1]).all()



import pandas as pd

from .const import COLS, COL_NAME


def annotation_to_pandas(data: list) -> pd.DataFrame:
    """
    Convert list of paths to a pandas table with coordinates.

    Parameters
    ----------
    data : list
        List of paths, each of which is a list of coordinates of shape N x 3,
            where N is the number of points in the path.

    Returns
    -------
    pd.DataFrame:
        pandas DataFrame with coordinates
    """

    df = pd.DataFrame()
    if len(data) > 0:
        for i, d in enumerate(data):
            cur_df = pd.DataFrame(d, columns=COLS)
            cur_df[COL_NAME] = i
            df = pd.concat([df, cur_df], ignore_index=True)
    return df


def pandas_to_annotations(df: pd.DataFrame) -> list:
    """
    Convert pandas table with coordinate to a list of paths.

    Parameters
    ----------
    df : pd.DataFrame
        pandas DataFrame with coordinates, including an ID column for individual paths.

    Returns
    -------
    list:
        List of paths, each of shape N x 3
    """
    data = []
    labels = []
    if len(df) > 0:
        for s in df[COL_NAME].unique():
            d = df[df[COL_NAME] == s][COLS].values
            data.append(d)
            labels.append(s)
    return data, labels



from __future__ import annotations

from skimage import io


def load_sample_image():
    img = io.imread("https://github.com/amedyukhina/napari-filament-annotator/raw/main/img/example_image.tif")
    return [(img, {})]



import numpy as np

from napari_filament_annotator.utils.geom import tetragon_intersection, compute_polygon_intersection


def test_tetragon_intersection(tetragons):
    p1, p2 = tetragons
    inter = tetragon_intersection(p1, p2)
    assert inter is not None
    assert isinstance(inter, np.ndarray)
    assert inter.shape == (2, 3)


def test_polygon_intersection(polygons):
    x = compute_polygon_intersection(polygons)
    assert x is not None
    assert isinstance(x, np.ndarray)
    assert len(x.shape) == 2
    assert x.shape[1] == 3



"""
Parameters for annotation display and refinement
"""

import json
import os

import numpy as np


class Params():

    def set_scale(self, scale):
        self.scale = np.array(scale)

    def set_smoothing(self, sigma_um):
        self.sigma = sigma_um / self.scale

    def set_linewidth(self, line_width):
        self.line_width = line_width

    def set_coef(self, alpha=0.01, beta=0.1, gamma=1):
        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma

    def set_ac_parameters(self, n_iter=100, n_interp=5, end_coef=0.01):
        self.n_iter = n_iter
        self.n_interp = n_interp
        self.end_coef = end_coef

    def save(self, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        params = dict(voxel_size_xy=self.scale[-1],
                      voxel_size_z=self.scale[0],
                      sigma_um=self.sigma[-1] * self.scale[-1],
                      line_width=self.line_width,
                      alpha=self.alpha,
                      beta=self.beta,
                      gamma=self.gamma,
                      n_iter=self.n_iter,
                      n_interp=self.n_interp,
                      end_coef=self.end_coef)
        with open(filename, 'w') as f:
            json.dump(params, f)



import pandas as pd
from napari_filament_annotator.utils.const import COL_NAME
from napari_filament_annotator.utils.io import annotation_to_pandas, pandas_to_annotations


def test_conversion(paths):
    df = annotation_to_pandas(paths)
    assert isinstance(df, pd.DataFrame)
    paths2, labels = pandas_to_annotations(df)
    assert isinstance(paths2, list)
    assert len(labels) == len(paths2) == len(paths) == len(df[COL_NAME].unique())
    assert sum([len(path) for path in paths2]) == len(df)
    assert sum([len(path) for path in paths]) == len(df)
    for i in range(len(paths)):
        assert (paths[i] == paths2[i]).all()



import os
import shutil
import tempfile

import numpy as np
import pytest
from scipy import ndimage


@pytest.fixture(scope='module')
def polygons():
    return [[[([0., 73., 12.]),
              ([0., 73., 25.]),
              ([0., 73., 38.]),
              ([0., 73., 55.]),
              ([0., 73., 69.])],
             [([10., 85., 12.]),
              ([10., 85., 25]),
              ([10., 85., 38]),
              ([10., 85., 55.]),
              ([10., 85., 69.])]],
            [[([0., 85., 15.]),
              ([0., 85., 29.]),
              ([0., 85., 41.]),
              ([0., 85., 54.]),
              ([0., 85., 61.]),
              ([0., 85., 72.])],
             [([10., 73., 15.]),
              ([10., 73., 29.]),
              ([10., 73., 41.]),
              ([10., 73., 54.]),
              ([10., 73., 61.]),
              ([10., 73., 72.])]]]


@pytest.fixture(scope='module')
def tetragons(polygons):
    npt1 = polygons[0][0]
    npt2 = polygons[1][0]
    fpt1 = polygons[0][1]
    fpt2 = polygons[1][1]

    p1 = [npt1[0], npt1[1], fpt1[1], fpt1[1]]
    p2 = [npt2[0], npt2[1], fpt2[1], fpt2[0]]
    return p1, p2


@pytest.fixture(scope='module')
def paths():
    n = np.random.randint(5, 10)
    paths = []
    for i in range(n):
        l = np.random.randint(10, 15)
        path = np.random.randint(0, 100, (l, 3))
        paths.append(path)
    return paths


@pytest.fixture(scope='module')
def img_snake():
    n = 100
    n2 = 50
    img = np.zeros((10, 50, 50))
    z = np.int_(np.round_(np.linspace(2, 8, n)))
    y = np.int_(np.round_(np.linspace(10, 30, n)))
    x = np.int_(np.round_(np.linspace(5, 40, n) + 10 * np.sin(np.linspace(0, 5, n))))
    img[z, y, x] = 1

    img = ndimage.gaussian_filter(img, 2)

    z2 = np.int_(np.round_(np.linspace(2, 8, n2)))
    y2 = np.int_(np.round_(np.linspace(10, 30, n2)))
    x2 = np.int_(np.round_(np.linspace(5, 40, n2) + 10 * np.sin(np.linspace(0, 5, n2))))
    return img, np.array([z2, y2, x2]).transpose()


@pytest.fixture(scope='module')
def tmp_path():
    path = tempfile.mkdtemp()
    os.makedirs(path, exist_ok=True)
    yield path
    shutil.rmtree(path)



import itertools

import numpy as np
from scipy import ndimage

from .utils.geom import compute_polygon_intersection
from .utils.postproc import snap_to_bright, gradient


class Annotator:
    """
    Annotator
    """

    def __init__(self, viewer, img_layer, params):
        self.params = params
        img = img_layer.data.copy().astype(np.float32)
        img = ndimage.gaussian_filter(img, sigma=params.sigma)
        self.grad = gradient(img, img_layer.scale)  # calculate the gradient for the active contour

        self.near_points = []  # store near points of the currently drawn polygon
        self.far_points = []  # store far points of the currently drawn polygon
        self.polygons = []  # store near and far points of the last 1-2 polygons (to compute intersections)
        self.annotation_layer = viewer.add_shapes(_get_bbox(img_layer.data.shape),
                                                  name='annotations',
                                                  shape_type='path',
                                                  edge_width=0,
                                                  scale=img_layer.scale,
                                                  blending='additive'
                                                  )
        self.viewer = viewer
        self.add_callbacks()

    def add_callbacks(self):
        self.annotation_layer.mouse_drag_callbacks.append(self._draw_polygon)
        self.annotation_layer.mouse_drag_callbacks.append(self._calculate_intersection)
        self.annotation_layer.bind_key('d', self.delete_the_last_shape)
        self.annotation_layer.bind_key('p', self.delete_the_last_point)
        self.annotation_layer.bind_key('f', self.delete_the_first_filament_point)
        self.annotation_layer.bind_key('l', self.delete_the_last_filament_point)

    def _draw_polygon(self, layer, event):
        """
        Draw a polygon under the current annotation curve.
        At each point of the annotation curve, we extend a line perpendicualar to the image view, and calculate
            the points intersection of this line with the near and far image border.
        These points of intersection - the near and far points - constitute the boundary of the polygon.
        """
        yield
        if 'Control' in event.modifiers:  # draw a polygon if "Control" is pressed
            # get the near and far points at the mouse position
            near_point, far_point = layer.get_ray_intersections(
                event.position,
                event.view_direction,
                event.dims_displayed
            )
            # append to the array of near and far points
            if (near_point is not None) and (far_point is not None):
                self.near_points.append(near_point)
                self.far_points.append(far_point)

            # draw a polygon from the array of near and far points
            if len(self.near_points) > 0 and len(self.far_points) > 0:
                self.draw_polygon(layer)

        yield

    def draw_polygon(self, layer, color: str = 'red'):
        """
        Draw a polygon between provided near and far points.

        Parameters
        ----------
        layer : napari.layers.Shapes
            napari shapes layer with annotations
        color : str, optional
            Color of the polygon

        Returns
        -------
        Updated shapes layer
        """
        near_points = self.near_points.copy()
        far_points = self.far_points.copy()
        if len(near_points) < 2:  # if only one point, add a temporary point for display purposes
            near_points.append(np.array(near_points[0]) + self.params.line_width)
            far_points.append(np.array(far_points[0]) + self.params.line_width)

        # reverse the far points and combine near and far points into a polygon
        far_points_reverse = far_points.copy()
        far_points_reverse.reverse()
        polygon = np.array(near_points + far_points_reverse)

        # remove the old polygon belonging to the same annotation
        if np.array((layer.data[-1][0] == polygon[0])).all():
            layer.selected_data = set(range(layer.nshapes - 1, layer.nshapes))
            layer.remove_selected()

        # add the updated polygon
        layer.add(
            polygon,
            shape_type='polygon',
            edge_width=self.params.line_width,
            edge_color=color
        )

    def _calculate_intersection(self, layer, event):
        """
        Calculate the intersection of two polygons, if the length of the polygons array == 2.
        """
        yield
        while event.type == 'mouse_move':
            if 'Control' not in event.modifiers:  # Only execute if "Control" is not pressed
                # add the last annotation to the polygon array and clear the annotation arrays
                self.calculate_intersection(layer)

            yield

    def calculate_intersection(self, layer):
        if len(self.near_points) > 0:
            self.polygons.append([self.near_points.copy(), self.far_points.copy()])
        self.near_points.clear()
        self.far_points.clear()

        # if there are 2 or more polygons, calculate their intersection
        if len(self.polygons) >= 2:
            filament = compute_polygon_intersection(self.polygons, layer.scale)
            filament = snap_to_bright(snake=filament, grad=self.grad,
                                      spacing=layer.scale, **vars(self.params))

            # remove the 2 polygons from the shapes layer
            layer.selected_data = set(range(layer.nshapes - 2, layer.nshapes))
            layer.remove_selected()

            # add the calculated filament
            layer.add(filament, shape_type='path', edge_color='green', edge_width=self.params.line_width)

            # clear the polygons array
            self.polygons.pop()
            self.polygons.pop()

    def delete_the_last_shape(self, layer, show_message=True):
        """
        Remove the last added shape (polygon or filament)

        """
        if layer.nshapes > 1:
            msg = 'delete the last added shape'

            # delete the last shape in the annotation layer
            layer.selected_data = set(range(layer.nshapes - 1, layer.nshapes))
            layer.remove_selected()

            if len(self.near_points) > 0:  # if any points in the near/far points array, clear them
                self.near_points.clear()
                self.far_points.clear()

            elif len(self.polygons) > 0:  # otherwise, clear the polygons array
                self.polygons.pop()
        else:
            msg = 'no shapes to delete'

        if show_message:
            layer.status = msg
            print(msg)

    def delete_the_last_point(self, layer):
        """
        Remove the last added point

        """
        if len(self.near_points) > 0 and len(self.far_points) > 0:
            self.near_points.pop()
            self.far_points.pop()
            if len(self.near_points) > 0:
                self.draw_polygon(layer)
            else:
                self.delete_the_last_shape(layer, show_message=False)

    def delete_the_last_filament_point(self, layer):
        """Remove the last point in the last filament"""
        if layer.nshapes > 1:
            data = layer.data[-1][:-1]
            layer.data = layer.data[:-1] + [data]

    def delete_the_first_filament_point(self, layer):
        """Remove the first point in the last filament"""
        if layer.nshapes > 1:
            data = layer.data[-1][1:]
            layer.data = layer.data[:-1] + [data]


def _get_bbox(shape):
    bbox = list(itertools.product(*[np.arange(2)
                                    for i in range(len(shape[-3:]))]))
    if len(shape) > 3:
        bbox = [(0,) + b for b in bbox]
    bbox = bbox * np.array(shape)
    return bbox



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"

from ._sample_data import load_sample_image
from ._widget import AnnotatorWidget

__all__ = (
    "load_sample_image",
    "AnnotatorWidget",
)



import numpy as np
from Geometry3D import *
from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist


def tetragon_intersection(p1, p2):
    """
    Calculate intersection of two tetragons in 3D

    Parameters
    ----------
    p1, p2 : list or array
        List of tetragon coordinates of shape 4x3

    Returns
    -------
    list or None:
        List of (two) coordinate of the intersection line or None if no intersection exists.
    """
    p1 = np.array(p1)
    p2 = np.array(p2)
    if p1.shape != (4, 3) or p2.shape != (4, 3):
        raise ValueError("Input polygon shape must be (4, 3)")
    p1 = list(set([Point(*coords) for coords in p1]))
    p2 = list(set([Point(*coords) for coords in p2]))

    if len(p1) > 2 and len(p2) > 2:
        plane1 = ConvexPolygon(p1)
        plane2 = ConvexPolygon(p2)
        inter = intersection(plane1, plane2)
        if inter is not None:
            inter = np.array([list(pt) for pt in inter])
        return inter
    else:
        return None


def compute_polygon_intersection(polygons, spacing=None):
    """
    Compute intersection of two polygons.

    Parameters
    ----------
    polygons : tuple
        Two polygons, each specified by a list of near and far points.
        Each polygon must have shape 2 x N x 3,
            the first row representing the near points, and the second row representing the far points,
            N is the number of points, 3 is the dimension of the image.
    spacing : tuple, list or array
        Voxel size, (z, y, x).

    Returns
    -------
    np.ndarray of shape M x 3
        List of M points for the polygon intersection.
    """
    # near and far points of the both polygons
    npt1 = polygons[0][0]
    npt2 = polygons[1][0]
    fpt1 = polygons[0][1]
    fpt2 = polygons[1][1]

    # convert spacing
    if spacing is None:
        spacing = [1, 1, 1]
    spacing = np.array(spacing)

    # calculate intersections for each pair of tetragons that constitute the provided polygons
    intersections = []
    for i in range(len(npt1) - 1):
        for j in range(len(npt2) - 1):
            p1 = [npt1[i], npt1[i + 1], fpt1[i + 1], fpt1[i]]
            p2 = [npt2[j], npt2[j + 1], fpt2[j + 1], fpt2[j]]
            inter = tetragon_intersection(p1, p2)
            if inter is not None:
                intersections.append(inter)
            else:
                intersections.append(np.ones([2, 3]) * -1)  # set to -1 if no intersection exists

    # select the largest intersections
    intersections = np.array(intersections).reshape(len(npt1) - 1, len(npt2) - 1, 2, 3)
    l = np.sqrt(np.sum((intersections[:, :, 0] - intersections[:, :, 1]) ** 2, -1))  # length of each intersection
    inds = linear_sum_assignment(l, maximize=True)  # match tetragons based on the intersection length
    overlap = intersections[inds[0], inds[1]]
    overlap = np.array([o for o in overlap if np.min(o) >= 0])  # remove the pairs with no intersection (the -1 values)

    # the start and end of each segment are sometimes swapped
    # we replace the start and end points by the center of each segment
    # we also add the start of the first segment and the end of the last segment
    # to find these, we calculate distances between each pair of points and select those that are furthest apart
    centers = [(overlap[i][0] + overlap[i][1]) / 2
               for i in range(1, len(overlap) - 1)]  # center of each intersection segment (except the start and end)
    overlap = np.array(overlap).reshape(-1, 3)
    dist = cdist(overlap * spacing, overlap * spacing)  # distance between each pair of points
    ind = np.where(dist == dist.max())[0]  # select indices of points that are the furthest apart
    overlap = np.array([overlap[ind[0]]] +
                       centers +
                       [overlap[ind[1]]])  # furtherst points + centers of remaining intersections
    return overlap
